#define FINALCOLOR_DISABLE_ALPHA_BLEND
#include "inc/windDeformer.hlsl"
#include "inc/CullMode.hlsl"

#include "../common/context.hlsl"
#include "../deferred/_HMD.hlsl"
#include "../deferred/_helper.hlsl"

#define COMPOSE_MASK float4(0, 1, 0, 1);

#define SHADOW_CULLING_NONE

#if !defined(EDGE) && defined(USE_DCS_DEFERRED)
	#include "common/stencil.hlsl"
	#define DCS_FinalColor_Stencil_Value STENCIL_COMPOSITION_FOLIAGE
#endif

#include "inc/DiffuseGainBiasBuildings.hlsl"
#include "inc/ApplyBuildingOptions.hlsl"
#include "terrain/inc/RENDER_VARYHOUSE_COMPONENTS.hlsl"

#include "inc/terrain.hlsl"
#include "../inc/PhotoshopBlends.hlsl"
#include "inc/Diffuse.hlsl"
#include "terrain/inc/normalToCamera.hlsl"

//----------------------------------------------------------------------------------------------------------------------------------------------------
// Под скользящими углами нет смысла рендерить вертикальный и горизонтальный билдборды : лезут артефакты, заметное "крестовое пересечение".
// Границы перекрытия билбордов, вне их рендериться только вертикальный или только горизонтальный.
const float billboardOffset = 0.3;
const float billboardVerticalPow = 8;
const float billboardHorizontalPow = 1.5;

#define ST_PI 3.1416
#define ST_PI2 (ST_PI*2.0)
#define TREE_ALPHA_TRESHOLD 0.1

#define THIS_IS_TREE
#define THIS_IS_VARYHOUSE_MATERIAL
//#define NOT_INVERT_CAST_MESH

//#define TEST_TREE_AO

// When rendering shadows camera pos is nonsence. But according to this issue https://jira.eagle.ru/browse/GS-71
// currently view matrix for shadows is broken (and so is TerrainContext_cameraDir), so we take camera pos during shadows rendering 
// assuming it is something close to validness.
// ... Or take sundir which should be equal to camera dir in view matrix
#define CAMERA_DIR_MODE 1

Texture2D material::diffuseTexture;
Texture2D material::paletteTexture;
Texture2D material::normalTexture;
int material::FORMATOFnormalTexture;

StructuredBuffer<float4> material::bilboardData = "syntheticTests/vertBB.sb";
float material::fade2bilboard = "0.0";
float material::verticalBilboards = "1.0";
float material::gamma = "1.0";
float material::windFlex = "0.01";

#ifdef NORMALS_FROM_CENTER
float4 material::center = "0, 10, 0, 0";
#endif

#ifdef TEST_TREE_AO
float3 material::height = "15.0";
#endif

//#define NORMALS_INTERPOLATE_UP
//----------------------------------------------------------------------------------------------------------------------------------------------------
float ApproxAcos(float x)
{
	return (-0.69813170079773212 * x * x - 0.87266462599716477) * x + 1.5707963267948966;
}

//----------------------------------------------------------------------------------------------------------------------------------------------------
float2 BillboardSelectMapFromAtlas(float2 vUnitTexCoords,		// texcoords of cutout billboard, defined in Compiler, in [0.0,1.0] range;
																// will be converted to correct smaller texcoords to access sub-image in atlas
	float3 vRightVector,		// right orientation of current instance
	float3 vOutVector,			// out orientation of current instance
	float3 vUpVector,			// up orientation of current instance
	float3 vCameraDirection)	// which way the main camera is facing
{
	// algorithmic depends on opposite direction
	float3 vDir = -vCameraDirection.xyz;

	// given a particular coordinate system some adjustments are necessary for algorithm that expects right-handed/Z-up
	vDir = float3(vDir.x, vDir.y, -vDir.z);

	// project camera direction onto plane that holds vRightVector and vOutVector vectors
	float fUpDot = dot(vDir, vUpVector);
	float3 vProjectedCamera = normalize(vDir - fUpDot * vUpVector);

	// determine which angle the projection lands on; corresponds to the correct bb sub-image
	float fRightDot = dot(vRightVector, vProjectedCamera);
	fRightDot = clamp(fRightDot, -1.0, 1.0);
	float fRightAngle = ApproxAcos(fRightDot);

	// need a value in [0,2pi] range, not [-pi,pi]
	float fOutDot = dot(vOutVector, vProjectedCamera);
	if (fOutDot < 0.0)
		fRightAngle = ST_PI2 - fRightAngle;

	// bump angle to be in the middle of the pie slice
	fRightAngle += ST_PI / material::verticalBilboards;
	fRightAngle = fmod(fRightAngle, ST_PI2);

	// convert angle to billboard image index
	int nImageIndex = int(fRightAngle / (ST_PI2 / material::verticalBilboards));

	float4 vTableEntry = material::bilboardData[nImageIndex];

	float2 vAtlasTexCoords = (vTableEntry.x < 0.0) ?
		float2(-vTableEntry.x + vTableEntry.z * vUnitTexCoords.y, vTableEntry.y + vTableEntry.w * (1 - vUnitTexCoords.x)) :
		float2(vTableEntry.x + vTableEntry.z * (1 - vUnitTexCoords.x), vTableEntry.y + vTableEntry.w * vUnitTexCoords.y);

	return vAtlasTexCoords;
}

//----------------------------------------------------------------------------------------------------------------------------------------------------
float distToCamera(float3 pos)
{
	float len = length(pos - TerrainContext_cameraPos - TerrainContext_Origin);
	return len;
}

float3x3 makeRotate(float3 from, float3 to)
{
	float3 v2 = normalize(from);
	float3 v1 = normalize(to);

	float3 l = cross(v1, v2);
	float cf = dot(v1, v2); // cos(fi)
	float df = 1. / (1. + cf);
	float3 m = l*df;
	float a = l.x*m.y;
	float b = l.y*m.z;
	float c = l.z*m.x;
	float3x3 M;
	M[0][0] = l.x*m.x + cf;
	M[1][1] = l.y*m.y + cf;
	M[2][2] = l.z*m.z + cf;
	M[0][1] = a - l.z;
	M[1][0] = a + l.z;
	M[0][2] = c + l.y;
	M[2][0] = c - l.y;
	M[1][2] = b - l.x;
	M[2][1] = b + l.x;
	return M;
}

//----------------------------------------------------------------------------------------------------------------------------------------------------
float4 @default_vs(
	TransformOutput transform,	
	// vertex streams
	float3 geometry::P,
	float3 geometry::N,
	int geometry::options,
	float2 geometry::diffuseTexCoord,
	float4 geometry::lodTransition,
	float1 geometry::geometryType,
	float3 geometry::tangent,
#ifdef TEST_TREE_AO
	float geometry::ambientOcclusion,
#endif

	// outputs
#ifdef NORMALS_FROM_CENTER
	out float3 center,
#endif
#ifdef TEST_TREE_AO
	out float1 ambientOcclusion,
#endif
	out float3 normal,
	out float3 tangent,
	out float3 binormal, 
	out float2 colorTexCoord,
	out float1 geometryType,
	out float1 blendFactor,
	out float1 normalFactor,
	out float1 seed,
	out float4 localpos
	)
{
	float4 position = float4(geometry::P, 1.0);
	normal = geometry::N;
	tangent = geometry::tangent;
#ifdef TEST_TREE_AO
	ambientOcclusion = geometry::ambientOcclusion;
#endif

	normalFactor = transform.blendFactor;
	ApplyBuildingOptions(transform, geometry::options);

	geometryType = geometry::geometryType;
	colorTexCoord = geometry::diffuseTexCoord;
	float lodDistance = transform.blendFactor;
	blendFactor = transform.blendFactor;

	float4 lodTransition = geometry::lodTransition;

#ifdef CAMERA_DIR_MODE
	#if CAMERA_DIR_MODE == 0
		float3 cameraToTree = -normalize(SharedParams_cameraPos.xyz - mul(float4(0, 0, 0, 1), transform.m).xyz);
	#elif CAMERA_DIR_MODE == 1
		float3 cameraToTree = TerrainContext_orthoCamera ? SharedParams_sunDirectionWS : -normalize(SharedParams_cameraPos.xyz - mul(float4(0, 0, 0, 1), transform.m).xyz);
	#endif
#else
	float3 cameraToTree = TerrainContext_orthoCamera ? TerrainContext_cameraDir : -normalize(SharedParams_cameraPos.xyz - mul(float4(0, 0, 0, 1), transform.m).xyz);
#endif

	if (geometryType < 0.5)
	{
		// branches
		position.xyz = lerp(lodTransition.xyz, position.xyz, lodDistance);
	}
	else if (geometryType < 1.5)
	{
		// fronds
		position.xyz = lerp(lodTransition.xyz, position.xyz, lodDistance);
	}
	else if (geometryType < 2.5)
	{
		// leafs
		position.xyz = lerp(lodTransition.xyz, position.xyz, lodDistance);
	}
	else if (geometryType < 3.5)
	{
		// facing leafs
		lodTransition.xy *= lerp(lodTransition.w, 1.0, lodDistance);
		float3 leafCorner = float3(-lodTransition.x, 0.0, lodTransition.y);
		
		float3x3 rotate = makeRotate(float3(0, 1, 0), cameraToTree);
		leafCorner = mul(leafCorner.xyz, rotate);

		position.xyz += leafCorner.xyz;
		position.xyz += lodTransition.z*cameraToTree;
	}
	else if (geometryType < 4.5)
	{
		// billboard vertical
		float3 vUp = float3(0.0, 1.0, 0.0);
		float3 vOut = -normalize(float3(cameraToTree.x, 0.0, cameraToTree.z));
		float3 vRight = normalize(cross(vUp, vOut));

		float3x3 mOrientation = float3x3(
			vRight.x, vOut.x, vUp.x,
			vRight.y, vOut.y, vUp.y,
			vRight.z, vOut.z, vUp.z);
		position.xyz = mul(mOrientation, position.xyz);
		normal = mul(mOrientation, float4(normal, 0)).xyz;
		tangent = mul(mOrientation, float4(tangent, 0)).xyz;

		colorTexCoord = BillboardSelectMapFromAtlas(float2(1-geometry::diffuseTexCoord.x, geometry::diffuseTexCoord.y),
			-normalize(transform.m[0].xyz),		// instance right
			-normalize(transform.m[2].xyz),		// instance out
			normalize(transform.m[1].xyz),		// instance up
			-cameraToTree);		// which way the main camera is facing

		transform.m = float4x4(
			length(transform.m[0]), 0, 0, 0,
			0, length(transform.m[1]), 0, 0,
			0, 0, length(transform.m[2]), 0,
			transform.m[3].x, transform.m[3].y, transform.m[3].z, 1
			);

#ifdef NORMALS_INTERPOLATE_UP		
		if ((transform.options & TOO_FARTRANSITION) == 0)
			normalFactor = 1.0;
#endif

		float angle = abs(dot(cameraToTree, float3(0, 1, 0)));
		position.xyz *= saturate(pow((2 - billboardOffset - angle), billboardVerticalPow));
	}
	else
	{
#ifdef NORMALS_INTERPOLATE_UP
		// billboard horisontal
		if ((transform.options & TOO_FARTRANSITION) == 0)
			normalFactor = 1.0;
#endif
		float angle = abs(dot(cameraToTree, float3(0, 1, 0)));
		position.xyz *= saturate(pow(angle + billboardOffset, billboardHorizontalPow));
	}

#ifdef NORMALS_FROM_CENTER
	center = mul(float4(material::center.xyz, 1), transform.m).xyz;
#endif

	localpos = position;
	position = mul(position, transform.m);

	if (geometryType < 3.5)
		ApplyWindDeformer(position, transform.m[3].xyz, geometryType < 0.5 ? 0.0 : 0.5, transform.seed + localpos.y, material::windFlex);

	normal = mul(float4(normal, 0), transform.m).xyz;
	normal = normalize(normal);
	tangent = mul(float4(tangent, 0), transform.m).xyz;
	tangent = normalize(tangent);
	binormal = cross(normal, tangent);

	seed = transform.seed;

	return position;
}

//----------------------------------------------------------------------------------------------------------------------------------------------------
#define DEBUG(color) {output.vDiffuse = color;return;};

void @decomposed_ps(
	inout ColorOutput output,
	inout float4 positionWS,
#ifdef NORMALS_FROM_CENTER
	float3 center,
#endif
#ifdef TEST_TREE_AO
	float1 ambientOcclusion,
#endif

	float3 normal,
	float3 tangent,
	float3 binormal,
	float2 colorTexCoord,
	float1 geometryType,
	float1 blendFactor,
	float1 normalFactor,
	float1 seed,
	float4 localpos
	)
{
	// normalmap
	//normal = normalToCamera(positionWS.xyz, normal);
	float3 ps_normal = normal;
	float4 surfacecolor = float4(1, 0, 0, 1);

	surfacecolor = material::diffuseTexture.Sample(WrapSampler, colorTexCoord);
	float4 normalTexColor = material::normalTexture.Sample(WrapSampler, colorTexCoord);
//	output.vDiffuse = normalTexColor.rgb; return;

	float3 normalmap = normalTexColor.xyz;
#ifdef NO_BC3N
	normalmap = NormalFromColor(normalTexColor, 0);
#else
	normalmap = NormalFromColor(normalTexColor, material::FORMATOFnormalTexture);
#endif

	float3x3 tangentSpace = float3x3(normalize(tangent), normalize(binormal), normalize(normal));
	ps_normal = mul(normalmap, tangentSpace);

	// not use pallete on frond, branches
	/*
	float paletted = geometryType < 1.5 ? 0.0 : 1.0;

	// palette
	float2 paletteUV = float2(seed, 0.75);
	float4 seedcolor = material::paletteTexture.SampleLevel(WrapSampler, paletteUV, 0);
	surfacecolor.rgb = lerp(surfacecolor.rgb, PhotoshopOverlay(surfacecolor.rgb, seedcolor.rgb), paletted);
	//surfacecolor.rgb = seedcolor.rgb;
	*/

	surfacecolor.rgb = pow(surfacecolor.rgb, 1 / material::gamma);
	//surfacecolor.r = blendFactor;
	//output.vDiffuse = seedcolor.rgb; return;

	/*
	if(geometryType < 0.5)
		surfacecolor.r = 1.0;
	else if (geometryType < 1.5)
		surfacecolor.g = 0.0;
	else if (geometryType < 2.5)
		surfacecolor.b = 1.0;
	else if (geometryType < 3.5)
		surfacecolor.rgb = 0.3;
	//*/

#ifdef NORMALS_INTERPOLATE_UP
	ps_normal = normalize(ps_normal);
	ps_normal = lerp(float3(0, 1, 0), ps_normal, geometryType > 3.5 ? normalFactor : 1.0);
#endif

	output.vDiffuse		= surfacecolor.rgb;

#ifdef NORMALS_FROM_CENTER
	if (geometryType < 3.5)
	{
		// mix sphere & source normals
		float3 coneNormal = -normalize(center - positionWS.xyz);
		float NoL = saturate(dot(coneNormal, SharedParams_sunDirectionWS.xyz) + 0.1);

		const float dotFactor = 0.6;
		float3 normal = lerp(coneNormal, ps_normal, NoL * dotFactor);

		ps_normal = lerp(ps_normal, normal, geometryType < 0.5/*leaves*/ ? 0.0 : 1.0);
		ps_normal = normalize(ps_normal);
	}
	else
	{
		output.vNormal = -normalize(center - positionWS.xyz);
	}
#endif
	output.vNormal		= normalize(ps_normal);

#ifdef PBR_TEXTURES
	output.fRoughness = TREE_ROUGHNESS;
	output.fMetallic = TREE_METALLIC;

#ifdef TEST_TREE_AO
	float fakeAO = (saturate(localpos.y / material::height) + 1.0)*0.5;
	output.fMetallic = fakeAO;
	//output.vDiffuse = float3(fakeAO, fakeAO, fakeAO);
#endif

#else
	//output.fDiffuseCorrection = material::normalCorrection;
	output.fSpecPower = 0.0f;
	output.fSpecFactor	= 0.0f;
#endif

#ifdef TEST_TREE_AO
	output.fAO = ambientOcclusion;
#endif
	output.vSelfIllumination = float4(0, 0, 0, 0);
	output.fReflectance	= 0.0f;
	output.fAlpha		= 1.0f;	
	output.fMaterialID  = TREE_MATERIAL;
	output.fHeightOverSurface = localpos.y;

#ifdef RENDER_VARYHOUSE_COMPONENTS
	if( SharedParams_shadingHint==RVC_SURFACE_COLOR)
		DEBUG(surfacecolor.rgb);
	if( SharedParams_shadingHint==RVC_FABRIC_COLOR)
		DEBUG(seed);
#endif

#if ENABLE_ATOC
  if (COVERAGE_MASK > 1)
  {
    surfacecolor.a -= material::fade2bilboard*(1 - blendFactor);
    output.fAlpha = (surfacecolor.a - TREE_ALPHA_TRESHOLD) / max(fwidth(surfacecolor.a), 0.0001) + 0.5;
    return;
  }
#endif
	// discard by alpha
	if( surfacecolor.a - material::fade2bilboard*(1 - blendFactor) < TREE_ALPHA_TRESHOLD)
		discard;
}

//----------------------------------------------------------------------------------------------------------------------------------------------------
float4 @flatShadows_vs(
	TransformOutput transform,	
	// vertex streams
	float3 geometry::P,
	float3 geometry::N,
	float2 geometry::diffuseTexCoord,
	float4 geometry::lodTransition,
	float1 geometry::geometryType,
	// outputs
	out float2 colorTexCoord,
	out float1 blendFactor
	)
{
	float4 position = float4(geometry::P, 1.0);

	colorTexCoord = geometry::diffuseTexCoord;
	float lodDistance = transform.blendFactor;
	blendFactor = transform.blendFactor;

	float4 lodTransition = geometry::lodTransition;

	float3 viewDir = normalize(SharedParams_sunDirectionWS);

	if (geometry::geometryType < 2.5)
	{
		// branches, fronds, leafs
		position.xyz = lerp(lodTransition.xyz, position.xyz, lodDistance);
	}
	else if (geometry::geometryType < 3.5)
	{
		// facing leafs
		lodTransition.xy *= lerp(lodTransition.w, 1.0, lodDistance);
		float3 leafCorner = float3(-lodTransition.x, 0.0, lodTransition.y);

		float3x3 rotate = makeRotate(float3(0, 1, 0), viewDir);
		leafCorner = mul(leafCorner.xyz, rotate);

		position.xyz += leafCorner.xyz;
		position.xyz += lodTransition.z*viewDir;
	}
	else if (geometry::geometryType < 4.5)
	{
		// billboard vertical
		float3 vUp = float3(0.0, 1.0, 0.0);
		float3 vOut = normalize(float3(viewDir.x, 0.0, viewDir.z));
		float3 vRight = normalize(cross(vUp, vOut));

		float3x3 mOrientation = float3x3(
			vRight.x, vOut.x, vUp.x,
			vRight.y, vOut.y, vUp.y,
			vRight.z, vOut.z, vUp.z);
		position.xyz = mul(mOrientation, position.xyz);

		colorTexCoord = BillboardSelectMapFromAtlas(geometry::diffuseTexCoord,
			-normalize(transform.m[0].xyz),		// instance right
			-normalize(transform.m[2].xyz),		// instance out
			normalize(transform.m[1].xyz),		// instance up
			viewDir);		// which way the main camera is facing

		transform.m = float4x4(
			length(transform.m[0]), 0, 0, 0,
			0, length(transform.m[1]), 0, 0,
			0, 0, length(transform.m[2]), 0,
			transform.m[3].x, transform.m[3].y, transform.m[3].z, 1
			);
	}
	else
	{
		// billboard horizontal
		float dotProd = dot(viewDir, float3(0.0, 1.0, 0.0));
		blendFactor *= saturate(dotProd + 0.3);
	}

	float localposY = position;
	position = mul(position, transform.m);

	if (geometry::geometryType < 3.5)
		ApplyWindDeformer(position, transform.m[3].xyz, geometry::geometryType < 0.5 ? 0.0 : 0.5, transform.seed + localposY, material::windFlex);

	return position;
}

//----------------------------------------------------------------------------------------------------------------------------------------------------
float @flatShadows_ps(
	inout float4 positionWS,
	float2 colorTexCoord,
	float1 blendFactor
	)
{
	float4 surfacecolor = float4(1, 0, 0, 1);

#if ENABLE_ATOC
  float alpha = 1;
  if (COVERAGE_MASK > 1)
  {
    surfacecolor = material::diffuseTexture.SampleBias(WrapSampler, colorTexCoord, 1.5);
    surfacecolor.a -= material::fade2bilboard*(1 - blendFactor);
    surfacecolor.a = (surfacecolor.a - TREE_ALPHA_TRESHOLD) / max(fwidth(surfacecolor.a), 0.0001) + 0.5;
    alpha = surfacecolor.a;
  }
  else
  {
    surfacecolor = material::diffuseTexture.Sample(WrapSampler, colorTexCoord);
    // discard by alpha
    if (surfacecolor.a - material::fade2bilboard*(1 - blendFactor) < TREE_ALPHA_TRESHOLD)
      discard;
  }
  return alpha;
#else
	surfacecolor = material::diffuseTexture.Sample(WrapSampler, colorTexCoord);

	// discard by alpha
	if (surfacecolor.a - material::fade2bilboard*(1 - blendFactor) < TREE_ALPHA_TRESHOLD)
		discard;

	return 1.0;
#endif
}

//----------------------------------------------------------------------------------------------------------------------------------------------------
void @reflection_ps(
	inout ReflectionOutput output,
	inout float4 positionWS,
	float3 normal,
	float2 colorTexCoord,
	float1 geometryType,
	float1 blendFactor
	)
{
	float2 distanceMaskTexCoord = colorTexCoord;

	// normalmap
	float3 ps_normal = normal;
	// normal from tangent
	// ...

	float4 surfacecolor = float4(1, 0, 0, 1);
	surfacecolor = material::diffuseTexture.Sample(WrapSampler, colorTexCoord);

	// TODO: add soft diffuse
	output.color.rgb = surfacecolor.rgb;
	output.color.a = 1;

	// discard by alpha
	if (surfacecolor.a - material::fade2bilboard*(1 - blendFactor) < TREE_ALPHA_TRESHOLD)
		discard;
}

//----------------------------------------------------------------------------------------------------------------------------------------------------
#include "inc/map.hlsl"

void @map_ps(
	inout MapOutput output,
	inout float4 positionWS,
	float2 colorTexCoord,
	float1 blendFactor

)
{
	float alpha = material::diffuseTexture.Sample(WrapSampler, colorTexCoord).a;

	// discard by alpha
	if (alpha - material::fade2bilboard*(1 - blendFactor) < TREE_ALPHA_TRESHOLD)
		discard;

	output.color = getMapColor(3);
}

//----------------------------------------------------------------------------------------------------------------------------------------------------
void @shadowmap_ps(
	inout float4 positionWS,
	float1 geometryType,
	float3 normal,
	float2 colorTexCoord,
	float1 blendFactor
	)
{
	float4 surfacecolor = float4(1, 0, 0, 1);
	surfacecolor = material::diffuseTexture.Sample(WrapSampler, colorTexCoord);

	// discard by alpha
	if (surfacecolor.a - material::fade2bilboard*(1 - blendFactor) < TREE_ALPHA_TRESHOLD)
		discard;
}


void @flir_ps(
	inout FlirOutput output,
	inout float4 positionWS,
#ifdef NORMALS_FROM_CENTER
	float3 center,
#endif
#ifdef TEST_TREE_AO
	float1 ambientOcclusion,
#endif
	float3 normal,
	float3 tangent,
	float3 binormal,
	float2 colorTexCoord,
	float1 geometryType,
	float1 blendFactor,
	float1 normalFactor,
	float1 seed,
	float4 localpos
)
{
	ColorOutput color;
	ColorOutput_Init(color);
	decomposed_ps(color, positionWS, 
#ifdef NORMALS_FROM_CENTER
		float3 center,
#endif
#ifdef TEST_TREE_AO
		0,
#endif
		normal, 
		tangent, 
		binormal, 
		colorTexCoord, 
		geometryType, 
		blendFactor, 
		normalFactor, 
		seed, 
		localpos);

	output.fDataType = geometryType < 0.5 ? TERRAIN_FLIR_TREE_BRANCHES : TERRAIN_FLIR_TREE_LEAVES;
	output.vDiffuse = color.vDiffuse;
	output.vNormal = color.vNormal;
	output.fAlpha = color.fAlpha;
}